{ Copyright (C) 1998-2000, written by Shkolnik Mike
  FIDOnet: 2:463/106.14
  E-Mail:  mshkolnik@scalabium.com
           mshkolnik@yahoo.com
  WEB: http://www.scalabium.com
       http://www.geocities.com/mshkolnik
  tel: 380-/44/-552-10-29

  The TSMLanguage component allows to develop the multilanguage application
  without recompile. You can define the external files with translated strings
  and custom messages and in run-time to load it for modification of component properties.
}
unit SMLang;

interface

uses Classes, IniFiles, TypInfo, SysUtils;

const
  chTabular = '~';

type
  TLanguageEvent = procedure(Sender: TObject; AComponent: TComponent;
                             PropertyName, OldValue: string; var NewValue: string) of object;
  TSMLanguage = class(TComponent)
  private
    FAutoTranslate: Boolean;
    FResourceFile: string;
    FProperties: TStrings;
    FSeparator: string;

    FOnBeforeTranslation: TNotifyEvent;
    FOnAfterTranslation: TNotifyEvent;
    FOnTranslate: TLanguageEvent;

    function GetFirstPart(var AString: string): string;
    function IsLastPart(AString: string): Boolean;
    function FindComponent(AOwner: TComponent; strName: string): TComponent;
    function GetProperty(AComponent: TComponent; strProperty: string): string;
    procedure SetProperty(AComponent: TComponent; strProperty, strValue: string);
    procedure TranslateProperty(strProperty, strTranslation: string);
    function GetStrings: TStrings;
    procedure SetStrings(Value: TStrings);
    procedure SetResourceFile(Value: string);
    procedure SetStringsProperty(AComponent: TComponent; PropInfo: PPropInfo; strValues: string);
  protected
    procedure SetSeparator(strSeparator: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SaveResources(AFileName: string);
    procedure LoadResources(AFileName: string);
    function TranslateCustom(strMessage: string): string;
  published
    property AutoTranslate: Boolean read FAutoTranslate write FAutoTranslate default False;
    property ResourceFile: string read FResourceFile write SetResourceFile;
    property Properties: TStrings read GetStrings write SetStrings;
    property Separator: string read FSeparator write SetSeparator;
    property OnBeforeTranslation: TNotifyEvent read FOnBeforeTranslation write FOnBeforeTranslation;
    property OnAfterTranslation: TNotifyEvent read FOnAfterTranslation write FOnAfterTranslation;
    property OnTranslate: TLanguageEvent read FOnTranslate write FOnTranslate;
  end;

implementation
uses Forms;

{ TSMLanguage }
constructor TSMLanguage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FProperties := TStringList.Create;
end;

destructor TSMLanguage.Destroy;
begin
  FProperties.Free;
  inherited Destroy;
end;

function TSMLanguage.GetFirstPart(var AString: string): string;
var i: Integer;
begin
  i := Pos('.', AString);
  if (i > 0) then
  begin
    Result := Copy(AString, 1, i-1);
    Delete(AString, 1, i);
  end
  else
  begin
    Result := AString;
    AString := '';
  end;
  AString := Trim(AString);
  Result := Trim(Result);
end;

function TSMLanguage.IsLastPart(AString: string): Boolean;
begin
  Result := (Pos('.', AString) = 0);
end;

function TSMLanguage.FindComponent(AOwner: TComponent; strName: string): TComponent;
var i: Integer;
begin
  Result := nil;
  strName := UpperCase(strName);
  if (AOwner <> nil) then
    with AOwner do
      for i := 0 to ComponentCount-1 do
        if (UpperCase(Components[i].Name) = strName) then
        begin
          Result := Components[i];
          break;
        end;
end;

function TSMLanguage.GetProperty(AComponent: TComponent; strProperty: string): string;
var PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(AComponent.ClassInfo, strProperty);
  if (PropInfo <> nil) then
    Result := GetStrProp(AComponent, PropInfo);
end;

procedure TSMLanguage.SetProperty(AComponent: TComponent; strProperty, strValue: string);
var
  PropInfo: PPropInfo;
begin
  if (AComponent <> nil) then
  begin
    PropInfo := GetPropInfo(AComponent.ClassInfo, strProperty);
    if (PropInfo <> nil) then
    begin
      case PropInfo^.PropType^^.Kind of
        tkString,
        tkLString,
        tkWString: SetStrProp(AComponent, PropInfo, strValue);
        tkInteger: SetOrdProp(AComponent, PropInfo, StrToInt(strValue))
      else
        SetStringsProperty(AComponent, PropInfo, strValue);
      end;
    end;
  end;
end;

procedure TSMLanguage.SaveResources(AFileName: string);
var
  resFile: TIniFile;

  procedure ProcessComponent(AComponent: TComponent);
  var
    i: Integer;
  begin
    for i := 0 to FProperties.Count-1 do
      if (Trim(FProperties[i]) <> '') then
        resFile.WriteString('Translations', FProperties[i], GetProperty(AComponent, FProperties[i]));

    for i := 0 to AComponent.ComponentCount-1 do
      ProcessComponent(AComponent.Components[i]);
  end;

begin
  if AFileName = '' then
    AFileName := ResourceFile;
  resFile := TIniFile.Create(AFileName);
  with resFile do
    try
      ProcessComponent(Self.Owner);
    finally
      Free;
    end;
end;

procedure TSMLanguage.LoadResources(AFileName: string);
var i: Integer;
begin
  if (Assigned(FOnBeforeTranslation)) then
    FOnBeforeTranslation(Self);
  if AFileName = '' then
    AFileName := ResourceFile;
  with TIniFile.Create(ResourceFile) do
    try
      for i := 0 to FProperties.Count-1 do
        if (Trim(FProperties[i]) <> '') then
          TranslateProperty(FProperties[i], ReadString('Properties', FProperties[i], ''));
    finally
      Free;
    end;
  if (Assigned(FOnAfterTranslation)) then
    FOnAfterTranslation(Self);
end;

procedure TSMLanguage.TranslateProperty(strProperty, strTranslation: string);
var
  AComponent: TComponent;
  i: Integer;
begin
  strProperty := Trim(strProperty);
  AComponent := Application;
  i := Pos(chTabular, strTranslation);
  while i > 0 do
  begin
    strTranslation := Copy(strTranslation, 1, i-1) + #9 + Copy(strTranslation, i+1, Length(strTranslation)-i);
    i := Pos(chTabular, strTranslation);
  end;
  if (strProperty <> '') then
    while (not IsLastPart(strProperty)) do
      AComponent := FindComponent(Self.Owner, GetFirstPart(strProperty))
  else
    exit;
  if ((AComponent <> nil) and (strTranslation <> '')) then
  begin
    if (Assigned(FOnTranslate)) then
      FOnTranslate(Self, AComponent, strProperty,
                   GetProperty(AComponent, strProperty), strTranslation);
    SetProperty(AComponent, GetFirstPart(strProperty), strTranslation);
  end;
end;

procedure TSMLanguage.SetStrings(Value: TStrings);
begin
  FProperties.Assign(Value);
end;

function TSMLanguage.GetStrings: TStrings;
begin
  Result := FProperties;
end;

procedure TSMLanguage.SetResourceFile(Value: string);
begin
  if Value <> FResourceFile then
  begin
    FResourceFile := Value;
    if (not (csLoading in ComponentState)) and FAutoTranslate then
      LoadResources('');
  end;
end;

function TSMLanguage.TranslateCustom(strMessage: string): string;
begin
  with TIniFile.Create(ResourceFile) do
    try
      Result := ReadString('Custom', strMessage, strMessage);
    finally
      Free;
    end;
end;

procedure TSMLanguage.SetStringsProperty(AComponent: TComponent;
                                         PropInfo: PPropInfo;
                                         strValues: string);
var
  AStrings: TStringList;
  strBuffer: string;
  i: Integer;
begin
  AStrings := TStringList.Create;
  i := Pos(FSeparator, strValues);
  while (i > 0) do
  begin
    strBuffer := Copy(strValues, 1, i-1);
    Delete(strValues, 1, i - 1 + Length(FSeparator));
    i := Pos(FSeparator, strValues);
    AStrings.Add(Trim(strBuffer));
  end;
  if (Length(Trim(strValues)) > 0) then
    AStrings.Add(Trim(strValues));
  SetOrdProp(AComponent, PropInfo, LongInt(Pointer(AStrings)));
  AStrings.Free;
end;

procedure TSMLanguage.SetSeparator(strSeparator: string);
begin
  if (Length(Trim(strSeparator)) = 0) then
    FSeparator := ','
  else
    FSeparator := strSeparator;
end;

end.
